/*
 Copyright (C) 2004  Scott McKellar  mck9@swbell.net

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 02111-1307, USA.

Implementation of Config
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "cfg.h"

struct CfgAttrAction;
typedef struct CfgAttrAction
{
	char attr_name[ TAG_LEN_MAX + 1 ];
	CfgActionType type;
} CfgAttrAction;

struct CfgAttrNode;
typedef struct CfgAttrNode
{
	struct CfgAttrNode * pNext;
	CfgAttrAction action;
} CfgAttrNode;

/* ----------------- */

struct CfgAction;
typedef struct CfgAction
{
	char tag[ TAG_LEN_MAX + 1 ];
	CfgActionType type;
	CfgAttrNode * first_attr;  /* linked list of attribute actions */
	CfgActionType default_attr_action;
} CfgAction;

struct CfgNode;
typedef struct CfgNode
{
	struct CfgNode * pNext;
	CfgAction action;
} CfgNode;

/* ----------------- */

#define CFG_MAGIC   1488598478UL
#define CFG_DEFUNCT 2298874841UL

typedef struct
{
	CfgNode * pFirst_action; /* unsorted singly-linked list */
	CfgActionType default_action;
	unsigned default_count;
	CfgActionType comment_action;
	unsigned comment_count;
	unsigned long magic;
} Cfg;

static CfgAction * seek_action( CfgNode * pNode, const char * tag );
static int add_action( Cfg * pCfg, const char * tag, CfgActionType type );
static void destruct_attr_list( CfgAttrNode ** ppNode );
static CfgAttrAction * seek_attr_action( CfgAttrNode * pNode,
	const char * name );
static int add_attr_action( Cfg * pCfg,
	const char * tag, const char * attr_name, CfgActionType type );
static int add_attr_default( Cfg * pCfg,
	const char * tag, CfgActionType type );

/****************************************************
 Note the action type for comments.
 ***************************************************/
int cfg_set_comment_action( Config config, CfgActionType type )
{
	int rc = 0;

	/* Sanity checks */

	if( ! cfg_valid( config ) )
	{
		fputs( "cfg_set_comment_action: no valid configuration supplied\n",
			stderr );
		rc = 1;
	}
	else
	{
		Cfg * pCfg = (Cfg *) config.p;

		if( pCfg->comment_count > 0 )
		{
			fputs( "cfg_set_comment_action: "
				"multiple comment actions specified\n", stderr );
			rc = 1;
		}

		++pCfg->comment_count;
		pCfg->comment_action = type;
	}

	return rc;
}

/****************************************************
 Note the action type for a specified tag.  This is
 mostly a sanity-checking wrapper for the internal
 function that does the real work.
 ***************************************************/
int cfg_set_tag_action( Config config, const char * tag,
	CfgActionType type )
{
	int rc = 0;

	/* Sanity checks */

	if( ! cfg_valid( config ) )
	{
		fputs( "cfg_set_tag_action: no valid configuration supplied\n",
			stderr );
		rc = 1;
	}
	else if( ! tag )
	{
		fputs( "cfg_set_tag_action: null ptr to tag\n", stderr );
		rc = 1;
	}
	else if( ! *tag )
	{
		fputs( "cfg_set_tag_action: tag name is empty\n", stderr );
		rc = 1;
	}
	else
	{
		switch( type )
		{
			case CFGA_KEEP :
			case CFGA_WARN :
			case CFGA_DROP :
			case CFGA_DROP_ALL :
			case CFGA_EDIT :
				rc = add_action( (Cfg *) config.p, tag, type );
				break;
			default :
				fprintf( stderr, "cfg_set_tag_action: "
					"Invalid action type for tag %s\n", tag );
				rc = 1;
				break;
		}
	}

	return rc;
}

/****************************************************
 For a specified tag, note the default attribute
 action.
****************************************************/
int cfg_set_attr_default( Config config, const char * tag,
	CfgActionType type )
{
	int rc = 0;

	/* Sanity checks */

	if( ! cfg_valid( config ) )
	{
		fputs( "cfg_set_attr_default: no valid configuration supplied\n",
			stderr );
		rc = 1;
	}
	else if( ! tag )
	{
		fputs( "cfg_set_attr_default: null ptr to tag\n", stderr );
		rc = 1;
	}
	else if( ! *tag )
	{
		fputs( "cfg_set_attr_default: tag name is empty\n", stderr );
		rc = 1;
	}
	else
	{
		switch( type )
		{
			case CFGA_KEEP :
			case CFGA_WARN :
			case CFGA_DROP :
				rc = add_attr_default( (Cfg *) config.p, tag, type );
				break;
			default :
				fprintf( stderr, "cfg_set_attr_default: "
					"Invalid action type for tag %s\n", tag );
				rc = 1;
				break;
		}
	}

	return rc;
}

/****************************************************
 Note the action type for a specified attribute of a
 specified tag.  This is mostly a sanity-checking
 wrapper for the internal function that does the real
 work.
 ***************************************************/
int cfg_set_attr_action( Config config, const char * tag,
	const char * attr_name, CfgActionType type )
{
	int rc = 0;

	/* Sanity checks */

	if( ! cfg_valid( config ) )
	{
		fputs( "cfg_set_attr_action: no valid configuration supplied\n",
			stderr );
		rc = 1;
	}
	else if( ! tag )
	{
		fputs( "cfg_set_attr_action: null ptr to tag\n", stderr );
		rc = 1;
	}
	else if( ! *tag )
	{
		fputs( "cfg_set_attr_action: tag name is empty\n", stderr );
		rc = 1;
	}
	else if( ! attr_name )
	{
		fprintf( stderr, "cfg_set_attr_action: "
			"null ptr to attribute name for %s tag\n", tag );
		rc = 1;
	}
	else if( ! *attr_name )
	{
		fprintf( stderr, "cfg_set_attr_action: attribute name "
			   "for %s tag is empty\n", tag );
		rc = 1;
	}
	else
	{
		switch( type )
		{
			case CFGA_KEEP :
			case CFGA_WARN :
			case CFGA_DROP :
				rc = add_attr_action( (Cfg *) config.p, tag,
						attr_name, type );
				break;
			default :
				fprintf( stderr, "cfg_set_attr_action: "
					"Invalid action type for tag %s\n", tag );
				rc = 1;
				break;
		}
	}

	return rc;
}

/****************************************************
 Assign a default action at the tag level.
 ***************************************************/
int cfg_set_tag_default( Config config, CfgActionType type )
{
	int rc = 0;

	/* Sanity checks */

	if( ! cfg_valid( config ) )
	{
		fputs( "cfg_set_tag_default: "
			"no valid configuration supplied\n", stderr );
		rc = 1;
	}
	else
	{
		Cfg * pCfg = (Cfg *) config.p;

		switch( type )
		{
			case CFGA_KEEP :
			case CFGA_WARN :
			case CFGA_DROP :
				pCfg->default_action = type;
				break;
			case CFGA_DROP_ALL :
				fputs( "cfg_set_tag_default: \"drop all\""
					"not allowed as default action\n", stderr );
				rc = 1;
				break;
			case CFGA_EDIT :
				fputs( "cfg_set_tag_default: \"edit\""
					"not allowed as default action\n", stderr );
				rc = 1;
				break;
			default :
				fputs( "cfg_set_tag_default: "
					"invalid default action\n", stderr );
				rc = 1;
				break;
		}

		if( pCfg->default_count > 0 )
		{
			fputs( "cfg_set_tag_default: "
				"multiple default actions specified\n", stderr );
			rc = 1;
		}

		++pCfg->default_count;
	}

	return rc;
}

/*************************************************
 Add an action of a specified type for a specified
 tag, to a specified configuration
 ************************************************/
static int add_action( Cfg * pCfg, const char * tag, CfgActionType type )
{
	int rc = 0;
	CfgNode * pNode = NULL;

	/* Skip most sanity checks; this is an internal */
	/* internal function and we trust ourselves     */

	/* check length of tag */

	if( strlen( tag ) > TAG_LEN_MAX )
	{
		fprintf( stderr, "add_action: overly long tag %s in "
				 "configuration file\n", tag );
		return 1;
	}

	/* See if the tag already exists in the list */

	if( seek_action( pCfg->pFirst_action, tag ) )
	{
		fprintf( stderr,
			"add_action: multiple actions specified for tag %s\n",
			tag );
		return 1;
	}

	/* Allocate new node, add it to the list */

	pNode = malloc( sizeof( *pNode ) );
	if( ! pNode )
	{
		fputs( "add_action: out of memory\n", stderr );
		rc = 1;
	}
	else
	{
		/* Initialize action.  Note that we assume     */
		/* that the tag has been raised to upper case. */

		strcpy( pNode->action.tag, tag );
		pNode->action.type = type;
		pNode->action.first_attr = NULL;
		pNode->action.default_attr_action = CFGA_KEEP;

		/* Add to head of list */

		pNode->pNext = pCfg->pFirst_action;
		pCfg->pFirst_action = pNode;
	}

	return rc;
}

/*************************************************
 Add an action of a specified type for a specified
 attribute, for a specified tag and configuration
 ************************************************/
static int add_attr_action( Cfg * pCfg,
	const char * tag, const char * attr_name, CfgActionType type )
{
	int rc = 0;
	CfgAction * pAction = NULL;
	CfgAttrNode * pAttrNode = NULL;

	/* Skip most sanity checks; this is an      */
	/* internal function and we trust ourselves */

	/* check length of attribute name */

	if( strlen( tag ) > TAG_LEN_MAX )
	{
		fprintf( stderr, "add_attr_action: overly long tag %s in "
				 "configuration file\n", tag );
		return 1;
	}

	/* Look for the tag in the tag list */

	pAction = seek_action( pCfg->pFirst_action, tag );
	if( NULL == pAction )
	{
		fprintf( stderr,
			"add_attr_action: tag %s does not exist\n", tag );
		return 1;
	}

	/* Make sure the attribute isn't already in the list */

	if( seek_attr_action( pAction->first_attr, attr_name ) )
	{
		fprintf( stderr,
			"add_atr_action: action for attribute %s already defined\n",
			attr_name );
		return 1;
	}

	/* Allocate new node, add it to the list */

	pAttrNode = malloc( sizeof( *pAttrNode ) );
	if( ! pAttrNode )
	{
		fputs( "add_attr_action: out of memory\n", stderr );
		rc = 1;
	}
	else
	{
		/* Initialize attribute action.  We assume that the */
		/* attribute name has been raised to upper case.    */

		strcpy( pAttrNode->action.attr_name, attr_name );
		pAttrNode->action.type = type;

		/* Add to head of list */

		pAttrNode->pNext = pAction->first_attr;
		pAction->first_attr = pAttrNode;
	}

	return rc;
}

/*************************************************
 Add an default attribute action for a specified
 tag and configuration
 ************************************************/
static int add_attr_default( Cfg * pCfg,
	const char * tag, CfgActionType type )
{
	int rc = 0;
	CfgAction * pAction = NULL;

	/* Skip most sanity checks; this is an      */
	/* internal function and we trust ourselves */

	/* Look for the tag in the tag list */

	pAction = seek_action( pCfg->pFirst_action, tag );
	if( NULL == pAction )
	{
		fprintf( stderr,
			"add_attr_default: tag %s does not exist\n", tag );
		rc = 1;
	}
	else
		pAction->default_attr_action = type;

	return rc;
}

/*************************************************
 Return the action type for comments
 ************************************************/
CfgActionType cfg_comment_action( Config config )
{
	Cfg * pCfg;
	static int reported = 0;   /* a boolean */

	if( ! cfg_valid( config ) )
	{
		/* Report error, unless we have already done so */

		if( ! reported )
		{
			fputs( "cfg_comment_action: invalid Config\n", stderr );
			reported = 1;
		}
		return CFGA_KEEP;
	}

	pCfg = (Cfg *) config.p;
	return pCfg->comment_action;
}

/*************************************************
 Determine the action type for a given tag
 ************************************************/
CfgActionType cfg_action( Config config, const char * tag )
{
	Cfg * pCfg;
	CfgAction * pAction;
	static int reported = 0;   /* a boolean */

	if( ! cfg_valid( config ) )
	{
		/* Report error, unless we have already done so */

		if( ! reported )
		{
			fputs( "cfg_action: invalid Config\n", stderr );
			reported = 1;
		}
		return CFGA_KEEP;
	}

	pCfg = (Cfg *) config.p;
	pAction = seek_action( pCfg->pFirst_action, tag );
	if( pAction )
		return pAction->type;
	else
		return pCfg->default_action;
}

/*************************************************
 Determine the action type for a given attribute
 of a given tag
 ************************************************/
CfgActionType cfg_attr_action( Config config, const char * tag,
	const char * attr )
{
	Cfg * pCfg;
	CfgAction * pAction;
	static int reported = 0;   /* a boolean */

	if( ! cfg_valid( config ) )
	{
		/* Report error, unless we have already done so */

		if( ! reported )
		{
			fputs( "cfg_action: invalid Config\n", stderr );
			reported = 1;
		}
		return CFGA_KEEP;
	}

	/* Look for the tag */

	pCfg = (Cfg *) config.p;
	pAction = seek_action( pCfg->pFirst_action, tag );
	if( pAction && pAction->first_attr )
	{
		/* We found the tag, now look for the attribute */

		CfgAttrAction * pAttrAction = seek_attr_action(
			pAction->first_attr, attr );

		if( pAttrAction )
			return pAttrAction->type;
		else
			return pAction->default_attr_action;
	}
	else
		return CFGA_KEEP; /* this shouldn't happen, but oh, well... */
}

/*********************************************************
 Traverse the list of actions, looking for a specified
 tag name.  If you find it, return a pointer to the
 action; otherwise return NULL.  Note that we rely on
 both tags to be in upper case, or at least the same case.
 *********************************************************/
static CfgAction * seek_action( CfgNode * pNode, const char * tag )
{
	while( pNode )
	{
		if( ! strcmp( tag, pNode->action.tag ) )
			return &(pNode->action);
		else
			pNode = pNode->pNext;
	}

	return NULL;
}

/*********************************************************
 Traverse the list of attribute actions, looking for a
 specified attribute tag name.  If you find it, return a
 pointer to the attribute action; otherwise return NULL.
 Note that we rely on both names to be in upper case, or
 at least the same case.
 *********************************************************/
static CfgAttrAction * seek_attr_action( CfgAttrNode * pNode,
	const char * name )
{
	while( pNode )
	{
		if( ! strcmp( name, pNode->action.attr_name ) )
			return &(pNode->action);
		else
			pNode = pNode->pNext;
	}

	return NULL;
}

/**************************************************
 Allocate and Initialize a Cfg
 *************************************************/
Config cfg_construct( void )
{
	Cfg * pCfg;
	Config cfg;

	pCfg = (Cfg *) malloc( sizeof *pCfg );
	if( pCfg )
	{
		pCfg->pFirst_action  = NULL;
		pCfg->default_action = CFGA_KEEP;
		pCfg->default_count  = 0;
		pCfg->comment_action = CFGA_KEEP;
		pCfg->comment_count  = 0;
		pCfg->magic          = CFG_MAGIC;
	}

	cfg.p = pCfg;
	return cfg;
}

/*******************************************
 Destruct a Config; deallocate all memory
 allocated to it
 ******************************************/
void cfg_destruct( Config * pConfig )
{
	if( pConfig && cfg_valid( *pConfig ) )
	{
		/* Deallocate all the nodes in the action list */

		Cfg * pCfg = (Cfg *) pConfig->p;
		CfgNode * pNode;
		CfgNode * pTemp;

		/* Free all the tag actions in the list -- and all the */
		/* attribute actions associated with each tag action.  */

		pNode = pCfg->pFirst_action;
		pCfg->pFirst_action = NULL;
		while( pNode )
		{
			pTemp = pNode->pNext;
			pNode->pNext = NULL;
			destruct_attr_list( &pNode->action.first_attr );
			free( pNode );
			pNode = pTemp;
		}

		pCfg->magic = CFG_DEFUNCT;
	}

	return;
}

/***************************************************
 Deallocate all the CfgAttrNodes in a list; nullify
 the list via a double pointer.
 **************************************************/
static void destruct_attr_list( CfgAttrNode ** ppNode )
{
	if( ppNode && *ppNode )
	{
		CfgAttrNode * pNode = *ppNode;
		CfgAttrNode * pTemp = NULL;

		while( pNode )
		{
			pTemp = pNode->pNext;
			pNode->pNext = NULL;
			free( pNode );
			pNode = pTemp;
		}
		*ppNode = NULL;
	}
}

/*****************************************
 Return 1 if the config is valid (so far
 as we can tell); otherwise return 0.
 ****************************************/
int cfg_valid( Config config )
{
	const Cfg * pCfg = (const Cfg *) config.p;

	if( pCfg && CFG_MAGIC == pCfg->magic )
		return 1;
	else
		return 0;
}
