/***************************************************
 Copyright (C) 2004  Scott McKellar  mck9@swbell.net

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 02111-1307, USA.

 Semantic analyzer for html_scrub configuration
 files.  It receieves a stream of tokens and parses
 them via a finite state machine.
 **************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "cfg.h"

typedef enum
{
	S_BETWEEN,
	S_LT,
	S_TAG,
	S_GT,
	S_T_COLON,
	S_T_DROP,
	S_T_LBRACE,
	S_T_DEFAULT,
	S_TD_COLON,    /* found colon after "default" at tag level */
	S_COMMENT,
	S_C_COLON,     /* found colon after "comment" */
	S_ATTRIBUTE,   /* found "attribute" starting attribute action */
	S_A_NAME,      /* found attribute name */
	S_A_COLON,     /* found colon after attribute name */
	S_A_DEFAULT,
	S_AD_COLON     /* found colon after "default" for attributes */
} State;

static State state;
static char tag_name[ CFG_TEXT_MAX + 1 ];
static char attribute_name[ CFG_TEXT_MAX + 1 ];
static Config config;

typedef int ( * State_handler )( const CfgTok * pTok );

static int do_between( const CfgTok * pT );
static int do_lt( const CfgTok * pT );
static int do_tag( const CfgTok * pT );
static int do_gt( const CfgTok * pT );
static int do_t_colon( const CfgTok * pT );
static int do_t_drop( const CfgTok * pT );
static int do_t_lbrace( const CfgTok * pT );
static int do_t_default( const CfgTok * pT );
static int do_td_colon( const CfgTok * pT );
static int do_comment( const CfgTok * pT );
static int do_c_colon( const CfgTok * pT );
static int do_attribute( const CfgTok * pT );
static int do_a_name( const CfgTok * pT );
static int do_a_colon( const CfgTok * pT );
static int do_a_default( const CfgTok * pT );
static int do_ad_colon( const CfgTok * pT );

static const State_handler state_handler[] =
{
	do_between,
	do_lt,
	do_tag,
	do_gt,
	do_t_colon,
	do_t_drop,
	do_t_lbrace,
	do_t_default,
	do_td_colon,
	do_comment,
	do_c_colon,
	do_attribute,
	do_a_name,
	do_a_colon,
	do_a_default,
	do_ad_colon
};

static FILE * pIn;
static int finished;  /* boolean */

/* keywords: */

static const char key_keep[]      = "KEEP";
static const char key_drop[]      = "DROP";
static const char key_warn[]      = "WARN";
static const char key_all[]       = "ALL";
static const char key_default[]   = "DEFAULT";
static const char key_comment[]   = "COMMENT";
static const char key_attribute[] = "ATTRIBUTE";

static void unexpected( const CfgTok * pT );
static const char * token_name( CfgTokType type );

int cfg_read( FILE * cfg_file, Config cfg )
{
	int rc = 0;
	CfgTok tok;

	/* sanity checks */

	if( NULL == cfg_file )
	{
		fputs( "cfg_read: Null pointer to configuration file\n",
			stderr );
		return 1;
	}
	else if( ! cfg_valid( cfg ) )
	{
		fputs( "cfg_read: invalid Config\n", stderr );
		return 1;
	}

	pIn = cfg_file;
	config = cfg;
	finished = 0;
	state = S_BETWEEN;

	/* Enter parsing loop */

	do
	{
		rc = cfg_tok( &tok, pIn );
		if( rc )
			fputs( "cfg_read: lexical error found\n", stderr );
		else if( T_ERROR == tok.type )
		{
			fprintf( stderr, "Error in configuration file: %s\n",
				tok.text );
			rc = 1;
		}
		else
		{
			rc = state_handler[ state ]( &tok );
			if( rc )
				fputs( "cfg_read: semantic error found\n", stderr );
		}

	} while ( ! finished && 0 == rc );

	config.p = NULL;
	pIn = NULL;
	return rc;
}

/*******************************************************
 We are between actions, and expect to see '<', the word
 "default", or end of file.
 ******************************************************/
static int do_between( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
			finished = 1;
			break;
		case T_LT :
			state = S_LT;
			break;
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_default ) )
				state = S_T_DEFAULT;
			else if( ! strcmp( pT->text, key_comment ) )
				state = S_COMMENT;
			else
			{
				unexpected( pT );
				rc = 1;
			}
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We just saw a '<', and now expect to see a tag name.
 ******************************************************/
static int do_lt( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			strcpy( tag_name, pT->text );
			state = S_TAG;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We just captured a tag name, and expect to see a '>'.
 ******************************************************/
static int do_tag( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
			unexpected( pT );
			rc = 1;
			break;
		case T_GT :
			state = S_GT;
			break;
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
		case T_WORD :
			unexpected( pT );
			rc = 1;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We just saw a '>', and expect to see a colon or a
 left curly brace.
*******************************************************/
static int do_gt( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
			unexpected( pT );
			rc = 1;
			break;
		case T_COLON :
			state = S_T_COLON;
			break;
		case T_LBRACE :
			rc = cfg_set_tag_action( config, tag_name, CFGA_EDIT );
			state = S_T_LBRACE;
			break;
		case T_RBRACE :
		case T_WORD :
			unexpected( pT );
			rc = 1;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found a tag and a colon.  We expect to see "keep",
 "drop", or "warn".
 ******************************************************/
static int do_t_colon( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
		case T_COLON :
			unexpected( pT );
			rc = 1;
			break;
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_drop ) )
				state = S_T_DROP;
			else if( ! strcmp( pT->text, key_keep ) )
			{
				rc = cfg_set_tag_action( config, tag_name, CFGA_KEEP );
				state = S_BETWEEN;
			}
			else if( ! strcmp( pT->text, key_warn ) )
			{
				rc = cfg_set_tag_action( config, tag_name, CFGA_WARN );
				state = S_BETWEEN;
			}
			else
			{
				unexpected( pT );   /* invalid word found */
				rc = 1;
				break;
			}
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found a tag, a colon, and "drop".  We expect to see
 either "all", the beginning of the next action, or
 end of file.
 ******************************************************/
static int do_t_drop( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
			rc = cfg_set_tag_action( config, tag_name, CFGA_DROP );
			finished = 1;
			break;
		case T_LT :
			rc = cfg_set_tag_action( config, tag_name, CFGA_DROP );
			state = S_LT;
			break;
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_all ) )
			{
				rc = cfg_set_tag_action( config, tag_name, CFGA_DROP_ALL );
				state = S_BETWEEN;
			}
			else if( ! strcmp( pT->text, key_default ) )
			{
				rc = cfg_set_tag_action( config, tag_name, CFGA_DROP );
				state = S_T_DEFAULT;
			}
			else
			{
				unexpected( pT );   /* invalid word found */
				rc = 1;
			}
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found a left curly brace after a colon.  We expect
 to see an attribute-level action specification, starting
 with either "default" (not yet implemented) or
 "attribute".
 ******************************************************/
static int do_t_lbrace( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
			finished = 1;
			break;
		case T_LT :
		case T_GT :
		case T_COLON :
		case T_LBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_RBRACE :
			state = S_BETWEEN;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_attribute ) )
				state = S_ATTRIBUTE;
			else if( ! strcmp( pT->text, key_default ) )
				state = S_A_DEFAULT;
			else
			{
				unexpected( pT );
				rc = 1;
			}
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found "default" at the tag level.  We expect to
 see a colon next.
 ******************************************************/
static int do_t_default( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
			unexpected( pT );
			rc = 1;
			break;
		case T_COLON :
			state = S_TD_COLON;
			break;
		case T_LBRACE :
		case T_RBRACE :
		case T_WORD :
			unexpected( pT );
			rc = 1;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found a colon after "default".  We expect to see an
 action.
 ******************************************************/
static int do_td_colon( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_drop ) )
			{
				rc = cfg_set_tag_default( config, CFGA_DROP );
				state = S_BETWEEN;
			}
			else if( ! strcmp( pT->text, key_warn ) )
			{
				rc = cfg_set_tag_default( config, CFGA_WARN );
				state = S_BETWEEN;
			}
			else if( ! strcmp( pT->text, key_keep ) )
			{
				rc = cfg_set_tag_default( config, CFGA_KEEP );
				state = S_BETWEEN;
			}
			else
			{
				unexpected( pT );
				rc = 1;
			}

			/* to do: add default action to list */

			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found "comment" at the tag level.  We expect to
 see a colon next.
 ******************************************************/
static int do_comment( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
			unexpected( pT );
			rc = 1;
			break;
		case T_COLON :
			state = S_C_COLON;
			break;
		case T_LBRACE :
		case T_RBRACE :
		case T_WORD :
			unexpected( pT );
			rc = 1;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found a colon after "comment".  We expect to see an
 action.
 ******************************************************/
static int do_c_colon( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_drop ) )
			{
				rc = cfg_set_comment_action( config, CFGA_DROP );
				state = S_BETWEEN;
			}
			else if( ! strcmp( pT->text, key_warn ) )
			{
				rc = cfg_set_comment_action( config, CFGA_WARN );
				state = S_BETWEEN;
			}
			else if( ! strcmp( pT->text, key_keep ) )
			{
				rc = cfg_set_comment_action( config, CFGA_KEEP );
				state = S_BETWEEN;
			}
			else
			{
				unexpected( pT );
				rc = 1;
			}

			/* to do: add default action to list */

			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found "attribute" at the beginning of an attribute
 action.  We expect the next token to be an attribute
 name.
 ******************************************************/
static int do_attribute( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			strcpy( attribute_name, pT->text );
			state = S_A_NAME;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found an attribute name.  We expect to see a colon.
 ******************************************************/
static int do_a_name( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
			unexpected( pT );
			rc = 1;
			break;
		case T_COLON :
			state = S_A_COLON;
			break;
		case T_LBRACE :
		case T_RBRACE :
		case T_WORD :
			unexpected( pT );
			rc = 1;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found a colon in an attribute action.  We expect
 to see KEEP, DROP, or WARN.
 ******************************************************/
static int do_a_colon( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_keep ) )
			{
				/* record KEEP action */

				rc = cfg_set_attr_action( config, tag_name,
						 attribute_name, CFGA_KEEP );
				state = S_T_LBRACE;
			}
			else if( ! strcmp( pT->text, key_drop ) )
			{
				/* record DROP action */

				rc = cfg_set_attr_action( config, tag_name,
						 attribute_name, CFGA_DROP );
				state = S_T_LBRACE;
			}
			else if( ! strcmp( pT->text, key_warn ) )
			{
				/* to do: record WARN action */

				rc = cfg_set_attr_action( config, tag_name,
						 attribute_name, CFGA_WARN );
				state = S_T_LBRACE;
			}
			else
			{
				unexpected( pT );   /* invalid word found */
				rc = 1;
				break;
			}
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We have found "default" starting an attribute action.
 We expect to see a colon.
 ******************************************************/
static int do_a_default( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
			unexpected( pT );
			rc = 1;
			break;
		case T_COLON :
			state = S_AD_COLON;
			break;
		case T_LBRACE :
		case T_RBRACE :
		case T_WORD :
			unexpected( pT );
			rc = 1;
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

/*******************************************************
 We found a colon after "default" in an attribute action.
 We expect to see KEEP, DROP, or WARN.
 ******************************************************/
static int do_ad_colon( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
		case T_LT :
		case T_GT :
		case T_COLON :
		case T_LBRACE :
		case T_RBRACE :
			unexpected( pT );
			rc = 1;
			break;
		case T_WORD :
			if( ! strcmp( pT->text, key_keep ) )
			{
				/* record KEEP default */;

				rc = cfg_set_attr_default( config, tag_name, CFGA_KEEP );
				state = S_T_LBRACE;
			}
			else if( ! strcmp( pT->text, key_drop ) )
			{
				/* record DROP default */;

				rc = cfg_set_attr_default( config, tag_name, CFGA_DROP );
				state = S_T_LBRACE;
			}
			else if( ! strcmp( pT->text, key_warn ) )
			{
				/* record WARN default */;

				rc = cfg_set_attr_default( config, tag_name, CFGA_WARN );
				state = S_T_LBRACE;
			}
			else
			{
				unexpected( pT );   /* invalid word found */
				rc = 1;
				break;
			}
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

#if 0

/* This function is a template for state handlers */

/*******************************************************
 ******************************************************/
static int do_( const CfgTok * pT )
{
	int rc = 0;

	switch( pT->type )
	{
		case T_EOF :
			finished = 1;
			break;
		case T_LT :
			break;
		case T_GT :
			break;
		case T_COLON :
			break;
		case T_LBRACE :
			break;
		case T_RBRACE :
			break;
		case T_WORD :
			break;
		default :
			fprintf( stderr, "Unrecognized token type %d\n",
				(int) pT->type );
			rc = 1;
			break;
	}

	return rc;
}

#endif

static void unexpected( const CfgTok * pT )
{
	if( T_WORD == pT->type )
		fprintf( stderr, "Unexpected word \"%s\" found\n", pT->text );
	else
		fprintf( stderr, "Unexpected %s found\n", token_name( pT->type ) );

	finished = 1;
}

static const char * token_name( CfgTokType type )
{
	switch( type )
	{
		case T_ERROR :
			return "error";
		case T_EOF :
			return "end of file";
		case T_LT :
			return "less-than symbol";
		case T_GT :
			return "greater-than symbol";
		case T_COLON :
			return "colon";
		case T_LBRACE :
			return "left curly brace";
		case T_RBRACE :
			return "right curly brace";
		case T_WORD :
			return "character string";
		default :
			return "unrecognized token type";
	}
}
