/****************************************************
 Copyright (C) 2004  Scott McKellar  mck9@swbell.net

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 02111-1307, USA.

 Tokenizer for configuration file.  Implements a
 small finite state machine.
 ***************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "cfg.h"

typedef enum
{
	S_BETWEEN,   /* between tokens   */
	S_WORD,      /* within a word    */
	S_COMMENT    /* within a comment */
} State;

static State state;
static size_t length;
static int eof_found;     /* boolean */
static CfgTok * pT;
static FILE * pIn;

static void do_between( int c );
static void do_word( int c );
static void do_comment( int c );

typedef void ( * State_handler )( int );

State_handler state_handler[] =
{
	do_between,
	do_word,
	do_comment
};

int cfg_tok( CfgTok * pTok, FILE * cfg_file )
{
	int rc = 0;
	int c;

	/* Sanity checks */

	if( NULL == pTok )
	{
		fprintf( stderr, "cfg_tok: Null pointer to token\n" );
		rc = 1;
	}
	if( NULL == cfg_file )
	{
		fprintf( stderr, "cfg_tok: Null file pointer\n" );
		rc = 1;
	}
	if( rc )
		return rc;

	/* initialize */

	eof_found = 0;
	pT = pTok;
	pIn = cfg_file;
	state = S_BETWEEN;

	/* enter loop of finite state machine */

	pTok->type = 0;
	do
	{
		errno = 0;
		c = getc( pIn );
		if( EOF == c && errno )
		{
				pTok->type = T_ERROR;
				strncpy( pTok->text, strerror( errno ), CFG_TEXT_MAX );
				pTok->text[ CFG_TEXT_MAX ] = '\0';
		}
		else
			state_handler[ state ]( c );

	} while( 0 == pTok->type );

	pIn = NULL;
	pT = NULL;
	return rc;
}

static void do_between( int c )
{
	switch( c )
	{
		case EOF :
			pT->type = T_EOF;
			pT->text[ 0 ] = '\0';
			eof_found = 1;
			break;
		case '<' :
			pT->type = T_LT;
			pT->text[ 0 ] = c;
			pT->text[ 1 ] = '\0';
			break;
		case '>' :
			pT->type = T_GT;
			pT->text[ 0 ] = c;
			pT->text[ 1 ] = '\0';
			break;
		case ':' :
			pT->type = T_COLON;
			pT->text[ 0 ] = c;
			pT->text[ 1 ] = '\0';
			break;
		case '{' :
			pT->type = T_LBRACE;
			pT->text[ 0 ] = c;
			pT->text[ 1 ] = '\0';
			break;
		case '}' :
			pT->type = T_RBRACE;
			pT->text[ 0 ] = c;
			pT->text[ 1 ] = '\0';
			break;
		case '#' :
			state = S_COMMENT;
			break;
		default :
			if( isspace( (unsigned char) c ) )
				;    /* ignore white space between tokens */
			else if( isalnum( (unsigned char) c ) )
			{
				pT->text[ 0 ] = toupper( (unsigned char) c );
				length = 1;
				state = S_WORD;
			}
			else
			{
				pT->type = T_ERROR;
				sprintf( pT->text, "Unexpected character %c", (char) c );
			}
			break;
	}
}

static void do_word( int c )
{
	if( isalnum( (unsigned char) c ) || '-' == c || '_' == c )
	{
		if( length < CFG_TEXT_MAX )
			pT->text[ length++ ] = toupper( (unsigned char) c );
		else
		{
			pT->type = T_ERROR;
			strcpy( pT->text, "Excessively long word encountered" );
		}
	}
	else if( EOF == c )
	{
		eof_found = 1;
		pT->type = T_WORD;
		pT->text[ length ] = '\0';
	}
	else
	{
		pT->type = T_WORD;
		pT->text[ length ] = '\0';
		state = S_BETWEEN;
		ungetc( c, pIn );  /* read it again next time */
	}
}

static void do_comment( int c )
{
	switch( c )
	{
		case '\n' :
			state = S_BETWEEN;
			break;
		case EOF :
			pT->type = T_EOF;
			pT->text[ 0 ] = '\0';
			eof_found = 1;
			break;
		default :
			break;
	}
}
