/*************************************************************
 Copyright (C) 2004  Scott McKellar  mck9@swbell.net

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 02111-1307, USA.

 This file is a collection of routines to implement an abstract
 data type Estr, or Expandable String.

 An Estr is a poor man's version of a string class for C programs.
 It stores a character string internally, and allocates additional
 memory as needed.  An Estr supports various operations on the
 underlying character string.  More may be added as needed.
*************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "estr.h"

#define OKAY 0
#define NOT_OKAY 1

/* Magic numbers with which to mark live or dead */
/* Imps, so that we can guard against invalid    */
/* pointers (though there is no guarantee)       */

#define ESTR_MAGIC 3148473736UL
#define DEFUNCT_MAGIC 2887409101UL

/* Minimum capacity of an Imp, in bytes; */
/* i.e. we don't create one any smaller  */

#define ESTR_MIN_CAPACITY 32

/* Implementation of an Elastic String */

typedef struct
{
	long magic;
	char * buff;
	size_t length;
	size_t capacity;
} Imp;

static void default_reporter( const char * s );

static Estr_error_reporter report = default_reporter;

static int expand( Imp * pimpl );
static int expand_n( Imp * pimpl, size_t new_capacity );

/*------------------------------------------------------*/
/* This is how we report an error, unless the client    */
/* code installs a different function to take its place */
/*------------------------------------------------------*/
static void default_reporter( const char * s )
{
	if( s )
	{
		fputs( s, stderr );
		fputc( '\n', stderr );
	}
}

/*------------------------------------------------------*/
/* Install a user-defined function for reporting errors */
/*------------------------------------------------------*/
Estr_error_reporter estr_install_error_reporter
( Estr_error_reporter pR )
{
	Estr_error_reporter pOldReporter = report;

	if( pR )
		report = pR;
	else
		report = default_reporter;

	return pOldReporter;
}

/*------------------------------------------------------*/
/* Return 1 if the Imp * appears to be valid, else      */
/* return 0.                                            */
/*                                                      */
/* We check for a null pointer and then check the magic */
/* number.                                              */
/*------------------------------------------------------*/
int estr_valid( Estr estr )
{
	Imp * pimpl = (Imp *) estr.p;

	if( NULL == pimpl )
	{
		report( "Estr is null" );
		return 0;
	}
	else if( ESTR_MAGIC == pimpl->magic )
		return 1;
	else
	{
		if( DEFUNCT_MAGIC == pimpl->magic )
			report( "Estr has been destroyed" );
		else
			report( "Estr is invalid" );

		return 0;
	}
}

/*------------------------------------------------------*/
/* Return the length of the stored data.                */
/*------------------------------------------------------*/
size_t estr_size( Estr estr )
{
	if( ! estr_valid( estr ) )
		return 0;
	else
	{
		Imp * pimpl = estr.p;
		return pimpl->length;
	}
}

/*------------------------------------------------------*/
/* Render an Estr empty.                                */
/*------------------------------------------------------*/
void estr_clear( Estr estr )
{
	if( ! estr_valid( estr ) )
		return;
	else
	{
		Imp * pimpl = estr.p;
		pimpl->length = 0;
	}
}

/*------------------------------------------------------*/
/* Construct an Imp with a mimimal initial capacity,    */
/* and return an Estr pointing to it.                   */
/*------------------------------------------------------*/
Estr estr_construct( void )
{
	return estr_construct_n( ESTR_MIN_CAPACITY );
}

/*------------------------------------------------------*/
/* Construct an Imp with a specified initial capacity,  */
/* and return an Estr pointing to it.                   */
/*------------------------------------------------------*/
Estr estr_construct_n( size_t n )
{
	Estr return_val;
	Imp * pimpl = NULL;

	if( n < ESTR_MIN_CAPACITY )
		n = ESTR_MIN_CAPACITY;

	pimpl = malloc( sizeof ( Imp ) );
	if( ! pimpl )
		report( "estr_construct_n: out of memory" );
	else
	{
		/* Allocate a buffer; initialize the Imp */

		pimpl->buff = malloc( n );
		if( pimpl->buff )
		{
			pimpl->magic    = ESTR_MAGIC;
			pimpl->length   = 0;
			pimpl->capacity = n;
		}
		else
		{
			report( "estr_construct_n: out of memory" );
			free( pimpl );
			pimpl = NULL;
		}
	}

	return_val.p = pimpl;
	return return_val;
}

/*------------------------------------------------------*/
/* Construct an Imp initialized from a nul-terminated   */
/* string and return an Estr pointing to it.            */
/*------------------------------------------------------*/
Estr estr_construct_str( const char * s )
{
	if( NULL == s || '\0' == *s )
		return estr_construct_n( ESTR_MIN_CAPACITY );
	else
		return estr_construct_mem( s, strlen( s ) );
}

/*------------------------------------------------------*/
/* Construct an Imp from a specified memory buffer, and */
/* return an Estr pointing to it.                       */
/*------------------------------------------------------*/
Estr estr_construct_mem( const void * p, size_t n )
{
	Estr estr;

	if( 0 == n || NULL == p || '\0' == *(const char *) p )
		return estr_construct_n( ESTR_MIN_CAPACITY );

	estr = estr_construct_n( n );
	if( estr.p )
	{
		Imp * pimpl = ( Imp * ) estr.p;
		memcpy( pimpl->buff, p, n );
		pimpl->length = n;
	}

	return estr;
}

/*------------------------------------------------------*/
/* Deallocate the Imp and its buffer.  Mark the Imp as  */
/* defunct so that we have some hope of detecting an    */
/* attempt to reuse it.                                 */
/*------------------------------------------------------*/
void estr_destruct( Estr * pEstr )
{
	Imp * pimpl;

	if( NULL == pEstr )
		return;

	pimpl = (Imp *) pEstr->p;

	if( pimpl ) /* to do: validate pimpl */
	{
		if( pimpl->buff )
			free( pimpl->buff );

		pimpl->magic = DEFUNCT_MAGIC; /* mark the Imp as defunct */
		free( pimpl );
	}

	pEstr->p = NULL;
}

/*------------------------------------------------------*/
/* Copy a specified null-terminated string to the       */
/* contents of the Estr, adjusting the length as needed */
/*------------------------------------------------------*/
int estr_assign_str( Estr estr, const char * s )
{
	if( ! estr_valid( estr ) )
		return NOT_OKAY;
	else
	{
		Imp * pimpl = estr.p;
		size_t length;

		if( ! s || ! *s )
		{
			/* Assign empty string */

			pimpl->length = 0;
			return OKAY;
		}

		length = strlen( s );
		if( length > pimpl->length )
		{
			/* Expand to accommodate string.  Note that this */
			/* expansion involves a needless copying of the  */
			/* buffer, which we will immediately overwrite.  */
			/* Here lies an opportunity for optimization.    */

			size_t new_capacity = length + length / 2;
			if( new_capacity < length )
				new_capacity = length;  /* avoid overflow */

			if( expand_n( pimpl, new_capacity ) &&
				expand_n( pimpl, length ) )
				return NOT_OKAY;
		}

		memcpy( pimpl->buff, s, length );
		pimpl->length = length;
		return OKAY;
	 }
}

/*------------------------------------------------------*/
/* Add a terminal nul and return a pointer to the       */
/* buffer.                                              */
/*------------------------------------------------------*/
const char * estr_c_str( Estr estr )
{
	/* Append a terminal nul */

	if( estr_append_c( estr, '\0' ) )
		return NULL;
	else
	{
		Imp * pimpl = estr.p;

		--pimpl->length;  /* don't include nul in length */
		return pimpl->buff;
	}
}

/*------------------------------------------------------*/
/* Append a single character, expanding the buffer as   */
/* needed.                                              */
/*------------------------------------------------------*/
int estr_append_c( Estr estr, char c )
{
	if( estr_valid( estr ) )
	{
		Imp * pimpl = estr.p;

		/* Expand if necessary */

		if( pimpl->length >= pimpl->capacity )
		{
			/* Need to expand the buffer.  Try successively */
			/* smaller expansions until you either succeed  */
			/* or have to give up. */

			if( expand( pimpl ) &&
				expand_n( pimpl, pimpl->capacity + 128 ) &&
				expand_n( pimpl, pimpl->capacity + 1 ) )
					return NOT_OKAY;  /* can't expand */
		}

		/* Append the character */

		pimpl->buff[ pimpl->length++ ] = c;
		return OKAY;
	}
	else
		return NOT_OKAY;  /* Invalid Estr */
}

/*------------------------------------------------------*/
/* Append a nul-terminated string, expanding the buffer */
/* as needed.                                           */
/*------------------------------------------------------*/
int estr_append_str( Estr estr, const char * s )
{
	if( ! s || ! *s )
		return OKAY;          /* nothing to append */
	else if( ! estr_valid( estr ) )
		return NOT_OKAY;
	else
	{
		size_t addend_length = strlen( s );
		Imp * pimpl = estr.p;
		size_t new_length = pimpl->length + addend_length;

		if( new_length < addend_length )
		{
			/* Numeric overflow: new length is longer */
			/* than maximum value of size_t           */

			report( "estr_append_str: Cannot expand buffer" );
			return NOT_OKAY;
		}
		else if( new_length > pimpl->capacity )
		{
			/* Expand the buffer */

			size_t new_capacity = new_length + new_length / 2;
			if( new_capacity < new_length )

				/* Overflow; back off on size */

				new_capacity = new_length;

			if( expand_n( pimpl, new_capacity ) )
				return NOT_OKAY;
		}

		memcpy( pimpl->buff + pimpl->length, s, addend_length );
		pimpl->length = new_length;
		return OKAY;
	}
}

/*------------------------------------------------------*/
/* Append a specified number of bytes, expanding the    */
/* buffer as needed.                                    */
/*------------------------------------------------------*/
int estr_append_mem( Estr estr, const char * s, size_t n )
{
	if( ! s )
		return OKAY;          /* nothing to append */
	else if( ! estr_valid( estr ) )
		return NOT_OKAY;
	else
	{
		Imp * pimpl = estr.p;
		size_t new_length = pimpl->length + n;

		if( new_length < n )
		{
			/* Numeric overflow: new length is longer */
			/* than maximum value of size_t           */

			report( "estr_append_mem: Cannot expand buffer" );
			return NOT_OKAY;
		}
		else if( new_length > pimpl->capacity )
		{
			/* Expand the buffer */

			size_t new_capacity = new_length + new_length / 2;
			if( new_capacity < new_length )

				/* Overflow; back off on size */

				new_capacity = new_length;

			if( expand_n( pimpl, new_capacity ) )
				return NOT_OKAY;
		}

		memcpy( pimpl->buff + pimpl->length, s, n );
		pimpl->length = new_length;
		return OKAY;
	}
}

/*------------------------------------------------------*/
/* Replace the existing buffer with a bigger one.  If   */
/* unsuccessful, leave the Imp unchanged.               */
/*                                                      */
/* We expand the buffer by a constant factor (50%) so   */
/* that the buffer grows in logarithmic time rather     */
/* than linear time.                                    */
/*                                                      */
/* This function is the heart of this entire source     */
/* file.                                                */
/*------------------------------------------------------*/
static int expand( Imp * pimpl )
{
	/* We don't check for a null pointer or an invalid Imp. */
	/* Since this is an internal function, we can trust     */
	/* ourselves to use it properly. */

	size_t new_capacity = pimpl->capacity;

	/* Calculate a new capacity, and check for overflow */

	new_capacity += new_capacity / 2;
	if( new_capacity < pimpl->capacity )
	{
		/* New size is too big.  If we knew the maximum value  */
		/* of a size_t we could set the length to that value,  */
		/* but we don't.  A size_t can be an int or a long,    */
		/* which in some implementations have different sizes. */

		report( "expand: unable to extend Estr buffer" );
		return NOT_OKAY;
	}
	else
		return expand_n( pimpl, new_capacity );
}

/*------------------------------------------------------*/
/* Replace the existing buffer with a bigger one of a   */
/* specified size.  If unsuccessful, leave the Imp      */
/* unchanged.                                           */
/*                                                      */
/* This function is the heart of this entire source     */
/* file.                                                */
/*------------------------------------------------------*/
static int expand_n( Imp * pimpl, size_t new_capacity )
{
	/* We don't check for a null pointer an invalid Imp, or */
	/* an implausible requested capacity.  Since this is an */
	/* internal function, we can trust ourselves to use it  */
	/* properly. */

	char * new_buff = malloc( new_capacity );
	if( ! new_buff )
	{
		report( "expand_n: out of memory for expanding Estr" );
		return NOT_OKAY;
	}

	/* Copy the contents to the new buffer.  Note that we could */
	/* have used realloc() instead of explicitly allocating,    */
	/* copying, and deallocating.  However, realloc would copy  */
	/* the entire block of memory, including any portion that   */
	/* isn't used.  Hence we arguably gain a minor efficiency   */
	/* by coding these steps ourselves. */

	memcpy( new_buff, pimpl->buff, pimpl->length );
	free( pimpl->buff );
	pimpl->buff = new_buff;
	pimpl->capacity = new_capacity;

	return OKAY;
}
