/********************************************************
main() for html_scrub.

 Copyright (C) 2004  Scott McKellar  mck9@swbell.net

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 02111-1307, USA.

html_scrub is a filter to eliminate certain HTML tags
and attributes from an HTML file, as specified by a
configuration file.  Error messages go to stderr, and
all other output goes to stdout.

In this source file we collect the command line arguments,
open files, and then delegate the real work to other routines.
 *******************************************************/
#include <stdlib.h>
#include <stdio.h>
#include "estr.h"
#include "cfg.h"
#include "hscrub.h"

#ifndef DEF_CFG_FILE
#define DEF_CFG_FILE "html_scrub.cfg"
#define DEF_DIR      "/etc"
#endif

static FILE * pCfg = NULL;
static FILE * pIn  = NULL;

static Config config;

static int get_args( int argc, char * argv[] );
static int default_cfg( void );

/*---------------------------------------------*/
/* Main routine                                */
/*---------------------------------------------*/
int main( int argc, char * argv[] )
{
	int rc;

	rc = get_args( argc, argv );

	if( 0 == rc && NULL == pCfg ) /* No configuration file      */
		rc = default_cfg();       /* specified?  Apply default. */

	if( 0 == rc )
	{
		/* Build the configuration from the configuration file */

		config = cfg_construct();
		if( ! cfg_valid( config ) )
		{
			fputs( "Unable to construct configuration\n", stderr );
			rc = 1;
		}
		else
		{
			install_config( config );

			/* read configuration file */

			rc = cfg_read( pCfg, config );
			if( rc != 0 )
				fputs( "Unable to interpret configuration file\n",
					   stderr );
		}
	}

	if( 0 == rc )
	{
		/* set up scrubbing functions */

		(void) install_tag_scrubber( scrub_tag );
		(void) install_text_scrubber( scrub_text );
		(void) install_comment_scrubber( scrub_comment );

		/* filter input file */

		rc = scrub_html( pIn, stdout );
	}

	/* Shut down */

	if( cfg_valid( config ) )
		cfg_destruct( &config );

	if( pCfg )
		fclose( pCfg );

	if( pIn && pIn != stdin )
		fclose( pIn );

	return rc;
}

/*-------------------------------------------------*/
/* Parse the command line.  Open the configuration */
/* file for input, if one is specified.  Also open */
/* the input file, if one is specified, or         */
/* designate stdin for use as input.               */
/*-------------------------------------------------*/
static int get_args( int argc, char * argv[] )
{
	const char * p;      /* ptr to argument */

	if( argc < 2 )
	{
		/* No arguments; apply default */

		pIn = stdin;
	}

	pCfg = pIn = NULL;
	p = *++argv;   /* point to first argument */

	/* Traverse the command line options */

	while( p && '-' == p[ 0 ] )
	{
		switch( p[ 1 ] )
		{
			case 'f'  :          /* Name of configuration file */
				if( pCfg )       /* If config file already found */
				{
					fputs( "Option -f may occur only once\n", stderr );
					return 1;
				}

				if( p[ 2 ] )
					p += 2;
				else
				{
					p = *++argv;
					if( ! p || ! *p || '-' == *p )
					{
						fputs( "No argument for -f option\n", stderr );
						return 1;
					}
				}

				/* Open the config file */

				pCfg = fopen( p, "r" );
				if( ! pCfg )
				{
					fprintf( stderr,
						"Unable to open configuration file %s\n", p );
					return 1;
				}

				break;
			case '\0' :   /* bare hyphen specifies stdin as input */
				pIn = stdin;
				break;
			default   :
				fprintf( stderr, "Unrecognized option %c on command line\n",
						(char) *p );
				return 1;
		}
		p = *++argv;
	}

	/* Look for an input file name on command line */

	if( p )
	{
		if( pIn )
		{
			fprintf( stderr, "Extraneous argument(s) on command line: %s\n",
					 p );
			return 1;
		}
		else if( ! p[ 0 ] )
		{
			fputs( "Name of specified input file is empty\n", stderr );
			return 1;
		}
		else
		{
			/* open the specified input file */

			pIn = fopen( p, "r" );
			if( ! pIn )
			{
				fprintf( stderr, "Unable to open input file %s\n", p );
				return 1;
			}
		}

		p = *++argv;
		if( p && p[ 0 ] )
		{
			fputs( "Extra argument(s) on command line\n", stderr );
			return 1;
		}
	}

	/* No input file specified?  Use standard input */

	if( ! pIn )
		pIn = stdin;

	return EXIT_SUCCESS;
}

/*---------------------------------------------*/
/* Open the configuration file.  Return 0 if   */
/* successful, otherwise 1.                    */
/*                                             */
/* First we look for the file name in an       */
/* environmental variable.  Then we try to     */
/* find the file in the current directory,     */
/* in the user's home directory, and in /etc.  */
/*---------------------------------------------*/
static int default_cfg( void )
{
	const char * fname = NULL;

	/* Try an environmental variable */

	fname = getenv( "HTML_SCRUB_CFG" );
	if( fname )
	{
		pCfg = fopen( fname, "r" );
		if( pCfg )
			return 0;
		else
		{
			fprintf( stderr,
				"Unable to open configuration file %s\n", fname );
			return 1;
		}
	}

	/* try the current directory */

	pCfg = fopen( DEF_CFG_FILE, "r" );
	if( pCfg )
		return 0;
	else
	{
		const char * home;	/* user's home directory */

		/* Create an expandable string for file names */

		Estr full_name = estr_construct();
		if( ! estr_valid( full_name ) )
		{
			fputs( "Unable to construct file name\n", stderr );
			return 1;
		}

		/* Look in user's home directory */

		home = getenv( "HOME" );
		if( home )
		{
			/* Build file name */

			if( estr_assign_str( full_name, home )
				|| estr_append_c( full_name, '/' )
				|| estr_append_str( full_name, DEF_CFG_FILE ) )
			{
				fputs( "Unable to allocate name for "
					"configuration file", stderr );
				estr_destruct( &full_name );
				return 1;
			}

			/* try to open the file */

			pCfg = fopen( estr_c_str( full_name ), "r" );
			if( pCfg )
			{
				estr_destruct( &full_name );
				return 0;
			}
		}

		/* That didn't work; try /etc */

		/* Build file name */

		if( estr_assign_str( full_name, DEF_DIR )
			|| estr_append_c( full_name, '/' )
			|| estr_append_str( full_name, DEF_CFG_FILE ) )
		{
			fputs( "Unable to allocate default name for "
				   "configuration file\n", stderr );
			estr_destruct( &full_name );
			return 1;
		}

		/* try to open file */

		pCfg = fopen( estr_c_str( full_name ), "r" );
		estr_destruct( &full_name );
		if( pCfg )
			return 0;
		else
		{
			fputs( "Unable to open a configuration file\n", stderr );
			return 1;
		}
	}
}
