/*----------------------------------------------------------
 Copyright (C) 2004  Scott McKellar  mck9@swbell.net

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 02111-1307, USA.

 Parse and filter an HTML file.  Pass the text unchanged,
 but invoke a scrubbing function (defined elsewhere) to
 delete or modify HTML tags and their attributes.
 ---------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "estr.h"
#include "cfg.h"
#include "hscrub.h"

/* We accumulate characters into a local buffer, and */
/* flush it as needed to an Estr (expandable string) */

#define MAX_LEN 127

static char buff [ MAX_LEN + 1 ];
static size_t len = 0;

static Estr estr;

static size_t start_line;
static size_t curr_line;

static int in_script = 0;  /* boolean */

/* States for a finite state machine */

typedef enum
{
	S_TEXT = 0,    /* In text, i.e. outside of tags */
	S_LESS,        /* Found '<' opening a tag */
	S_TAG,         /* Within '<' and '>' of a tag */
	S_QUOTE,       /* Within the quoted value of an attribute */
	S_BANG,        /* Found '!' opening a comment tag */
	S_DASH_1,      /* Found first hyphen opening a comment */
	S_DASH_2,      /* Found second hyphen opening a comment */
	S_DASH_3,      /* Found first hyphen closing a comment */
	S_ERROR,       /* Invalid syntax within a comment tag */
	S_SCRIPT_LESS  /* Found '<' within script text */
} State;

static int do_text( int c );
static int do_less( int c );
static int do_tag( int c );
static int do_quote( int c );
static int do_bang( int c );
static int do_dash_1( int c );
static int do_dash_2( int c );
static int do_dash_3( int c );
static int do_error( int c );
static int do_script_less( int c );

/* Declare an array of function */
/* pointers, one for each state */

typedef int ( * State_func )( int );

static State_func func_array[] =
{
	do_text,
	do_less,
	do_tag,
	do_quote,
	do_bang,
	do_dash_1,
	do_dash_2,
	do_dash_3,
	do_error,
	do_script_less
};

static State state;
static FILE * pOut;

static int write_buffer( void );
static int finish_buffer( void );
static int append_c( int c );
static int process_text( void );
static int process_comment( void );
static int process_tag( void );

static int default_text_scrubber( const char * s, FILE * f, size_t line );
static int put_character( char c );
static int default_tag_scrubber( const char * s, FILE * f, size_t line );

static void check_script( const char * s );

static Scrubber text_scrubber = default_text_scrubber;
static Scrubber tag_scrubber = default_tag_scrubber;
static Scrubber comment_scrubber = default_tag_scrubber;

/*---------------------------------------------*/
/* Three functions to install callback         */
/* functions.  Each returns a pointer to the   */
/* previous function.  Passing a NULL installs */
/* the default function.                       */
/*---------------------------------------------*/
Scrubber install_text_scrubber( Scrubber scrubber )
{
	Scrubber temp = text_scrubber;
	text_scrubber = scrubber ? scrubber : default_text_scrubber;
	return temp;
}

Scrubber install_tag_scrubber( Scrubber scrubber )
{
	Scrubber temp = tag_scrubber;
	tag_scrubber = scrubber ? scrubber : default_tag_scrubber;
	return temp;
}

Scrubber install_comment_scrubber( Scrubber scrubber )
{
	Scrubber temp = comment_scrubber;
	comment_scrubber = scrubber ? scrubber : default_tag_scrubber;
	return temp;
}

/*---------------------------------------------*/
/* Implement a small finite state automaton    */
/* for parsing HTML into tags, comments, and   */
/* text.                                       */
/*                                             */
/* We use a table of function pointers indexed */
/* by state.                                   */
/*---------------------------------------------*/
int scrub_html( FILE * pIn, FILE * pOutput )
{
	int rc = 0;
	int c;

	pOut = pOutput;
	estr = estr_construct();

	curr_line = start_line = 1;
	state = S_TEXT;

	do
	{
		c = getc( pIn );
		if( c )           /* silently suppress nuls */
		{
			rc = func_array[ state ]( c );
			if( '\n' == c )
				++curr_line;
		}
	}
	while( c != EOF && ! rc );

	estr_destruct( &estr );
	pOut = NULL;
	return rc;
}

/* Series of functions, one for each state... */

/*---------------------------------------------*/
/* We are in the text portion of the HTML,     */
/* i.e. outside of any angle brackets.         */
/*---------------------------------------------*/
static int do_text( int c )
{
	int rc = 0;

	switch( c )
	{
		case '<' :
			if( in_script )
				state = S_SCRIPT_LESS;
			else
			{
				rc = process_text();
				start_line = curr_line;
				state = S_LESS;
			}
			break;
		case EOF :
			rc = process_text();
			break;
		default  :
			rc = append_c( c );
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We have encountered a left angle bracket    */
/* ( '<' ), so we know we are in a tag, but we */
/* don't know yet whether it is a comment tag  */
/* (which requires some special treatment) or  */
/* a normal tag.                               */
/*---------------------------------------------*/
static int do_less( int c )
{
	int rc = 0;

	switch( c )
	{
		case ' ' :
		case '\t' :
		case '\n' :
			rc = append_c( c );
			break;
		case '!' :
			rc = append_c( c );
			state = S_BANG;
			break;
		case '>' :
			rc = process_tag();
			estr_clear( estr );
			len = 0;
			start_line = curr_line;
			state = S_TEXT;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			/* Now we know it's not a comment... */

			rc = append_c( c );
			state = S_TAG;
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* Upon entering a tag and skipping any white  */
/* space, we have encountered something other  */
/* than a bang ( '!' ).  So we know we are in  */
/* an ordinary tag rather than a comment tag.  */
/*---------------------------------------------*/
static int do_tag( int c )
{
	int rc = 0;

	switch( c )
	{
		case '"'  :
			rc = append_c( c );
			state = S_QUOTE;
			break;
		case '>' :
			rc = process_tag();
			estr_clear( estr );
			len = 0;
			start_line = curr_line;
			state = S_TEXT;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			rc = append_c( c );
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We are inside the quoted value of an        */
/* attribute.  Treat closing angle brackes as  */
/* literal characters, not as delimiters.      */
/*---------------------------------------------*/
static int do_quote( int c )
{
	int rc = 0;

	switch( c )
	{
		case '"' :
			rc = append_c( c );
			state = S_TAG;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			rc = append_c( c );
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We have found a bang ( '!' ) at the         */
/* beginning of a tag, so we know it's a       */
/* comment tag.                                */
/*                                             */
/* Alternatively, we have just found the end   */
/* of a comment.  Either way, we look for      */
/* either the closing angle bracket or the     */
/* beginning of a comment.                     */
/*                                             */
/* Apart from the opening and closing angle    */
/* brackets, a comment tag consists of the     */
/* bang itself followed by zero or more        */
/* comments.  There may be more than one       */
/* comment within a tag.                       */
/*                                             */
/* Each comment begins with a double-dash      */
/* ( "--" ) and ends with another double-dash. */
/* The text between the double-dashes may      */
/* contain any characters, as long as it does  */
/* not include double-dashes.                  */
/*                                             */
/* This syntax is not always understood by     */
/* those who write HTML.  If we detect a       */
/* syntax error, we go into an error state     */
/* where we merely collect all characters up   */
/* to the next closing angle bracket.          */
/*---------------------------------------------*/
static int do_bang( int c )
{
	int rc = 0;

	switch( c )
	{
		case ' ' :
		case '\t' :
		case '\n' :
			rc = append_c( c );
			break;
		case '-' :
			rc = append_c( c );
			state = S_DASH_1;
			break;
		case '>' :
			rc = process_comment();
			start_line = curr_line;
			state = S_TEXT;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			rc = append_c( c );
			state = S_ERROR;
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We have seen the first dash in the opening  */
/* of a comment.  We expect the next character */
/* to be a dash as well.                       */
/*---------------------------------------------*/
static int do_dash_1( int c )
{
	int rc = 0;

	switch( c )
	{
		case '-' :
			rc = append_c( c );
			state = S_DASH_2;
			break;
		case '>' :
			rc = process_comment();
			start_line = curr_line;
			state = S_TEXT;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			rc = append_c( c );
			state = S_ERROR;
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We have seen the second dash in the opening */
/* of a comment.  Collect every character, but */
/* change state when we see another dash.      */
/*---------------------------------------------*/
static int do_dash_2( int c )
{
	int rc = 0;

	switch( c )
	{
		case '-' :
			rc = append_c( c );
			state = S_DASH_3;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			rc = append_c( c );
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We have found a dash within a comment.  It  */
/* may or may not be the first dash of two.    */
/* if the next character is a dash, close the  */
/* comment.  Otherwise we will still be in the */
/* comment, so return to the appropriate state.*/
/*---------------------------------------------*/
static int do_dash_3( int c )
{
	int rc = 0;

	switch( c )
	{
		case '-' :
			rc = append_c( c );
			state = S_BANG;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			rc = append_c( c );
			state = S_DASH_2;
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We have found a syntax error within a       */
/* comment.  Keep accumulating characters      */
/* until a closing angle bracket turns up.     */
/*---------------------------------------------*/
static int do_error( int c )
{
	int rc = 0;

	switch( c )
	{
		case '>' :
			rc = process_comment();
			start_line = curr_line;
			state = S_TEXT;
			break;
		case EOF :
			fputc( '<', pOut );
			rc = write_buffer();
			break;
		default  :
			rc = append_c( c );
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* We have found a '<' within a script.  If    */
/* the next character is a virgule ('/'),      */
/* treat the '<' as the beginning of a tag;    */
/* otherwise treat it as part of the script.   */
/*---------------------------------------------*/
static int do_script_less( int c )
{
	int rc = 0;

	switch( c )
	{
		case '/' :  /* interpret as end tag */
			rc = process_text();
			start_line = curr_line;
			if( 0 == rc )
				rc = append_c( '/' );
			state = S_TAG;
			break;
		case '!' : /* interpret as start of HTML comment */
			rc = process_text();
			start_line = curr_line;
			if( 0 == rc )
				rc = append_c( c );
			state = S_BANG;
			break;
		case EOF :
			rc = process_text();
			start_line = curr_line;
			fputc( '<', pOut );
			break;
		default  :  /* interpret as script text */
			rc = append_c( '<' );
			rc = append_c( c );
			state = S_TEXT;
			break;
	}

	return rc;
}

/*---------------------------------------------*/
/* Append a character to the local buffer.  If */
/* the buffer is full, append the buffer to an */
/* Expandable String, then put the character   */
/* at the beginning of the buffer.             */
/*---------------------------------------------*/
static int append_c( int c )
{
	if( len < MAX_LEN )
		buff[ len++ ] = (char) c;
	else
	{
		/* flush buffer */

		if( finish_buffer() )
		{
			fputs( "Unable to allocate HTML buffer\n", stderr );
			return 1;
		}

		buff[ 0 ] = (char) c;
		len = 1;
	}

	return 0;
}

/*---------------------------------------------*/
/* Append the contents of the local buffer to  */
/* the Expandable String.                      */
/*---------------------------------------------*/
static int finish_buffer( void )
{
	if( len )
	{
		if( estr_append_mem( estr, buff, len ) )
		{
			fputs( "Unable to expand HTML buffer\n", stderr );
			return 1;
		}
		else
		{
			len = 0;
			return 0;
		}
	}
	else
		return 0;  /* nothing to append */
}

/*---------------------------------------------*/
/* Write the accumulated characters to the     */
/* current output file.                        */
/*---------------------------------------------*/
static int write_buffer( void )
{
	int rc = 0;

	if( finish_buffer() )
		rc = 1;
	else
	{
		const char * s = estr_c_str( estr );
		if( s && *s )
		{
			/* to do: check for IO errors */

			fputs( s, pOut );
			estr_clear( estr );
		}
	}
	return rc;
}

/*---------------------------------------------*/
/* Pass the accumulated text to the current    */
/* text scrubbing function.                    */
/*---------------------------------------------*/
static int process_text()
{
	int rc = 0;
	const char * s;

	if( finish_buffer() )
		return 1;

	s = estr_c_str( estr );
	if( s && *s )
		rc = text_scrubber( s, pOut, start_line );

	estr_clear( estr );

	return rc;
}

/*---------------------------------------------*/
/* Pass the accumulated comment to the current */
/* comment scrubbing function.                 */
/*---------------------------------------------*/
static int process_comment()
{
	int rc = 0;
	const char * s;

	if( finish_buffer() )
		return 1;

	s = estr_c_str( estr );
	if( s && *s )
		rc = comment_scrubber( s, pOut, start_line );

	estr_clear( estr );

	return rc;
}
/*---------------------------------------------*/
/* Pass the accumulated tag to the current tag */
/* scrubbing function.                         */
/*---------------------------------------------*/
static int process_tag()
{
	int rc = 0;
	const char * s;

	if( finish_buffer() )
		return 1;

	s = estr_c_str( estr );
	if( s && *s )
	{
		check_script( s );
		rc = tag_scrubber( s, pOut, start_line );
	}

	estr_clear( estr );

	return rc;
}

/*---------------------------------------------*/
/* Write the specified text to the specified   */
/* file, without alteration.  Ignore the line  */
/* number.                                     */
/*---------------------------------------------*/
static int default_text_scrubber( const char * s, FILE * f, size_t line )
{
	if( ! s || ! f )
		return 0;

	if( fputs( s, f ) == EOF )
	{
		perror( "Unable to write HTML" );
		return 1;
	}
	else
		return 0;
}

/*---------------------------------------------*/
/* Write a single character to the output      */
/* file.  In practice this will be a left or   */
/* right angle bracket.                        */
/*---------------------------------------------*/
static int put_character( char c )
{
	if( fputc( c, pOut ) == EOF )
	{
		perror( "Unable to write to HTML file" );
		return 1;
	}
	else
		return 0;
}

/*---------------------------------------------*/
/* Write the specified string within angle     */
/* brackets.                                   */
/*---------------------------------------------*/
static int default_tag_scrubber( const char * s, FILE * f, size_t line )
{
	if( ! s || ! f )
		return 0;

	if( put_character( '<' )
		|| default_text_scrubber( s, f, line )
		|| put_character( '>' ) )
		return 1;
	else
		return 0;
}

/*-----------------------------------------------*/
/* Check whether a tag is SCRIPT (or /SCRIPT)    */
/*-----------------------------------------------*/
static void check_script( const char * s )
{
	int ending = 0;  /* true if an end tag */

	/* fail-safe */

	if( !s || !*s )
		return;

	/* skip leading white space */

	while( isspace( (unsigned char) *s ) )
		++s;

	if( '/' == *s )
	{
		/* Found a virgule, denoting an end tag */

		ending = 1;
		++s;

		/* skip white space again */

		while( isspace( (unsigned char) *s ) )
			++s;
	}

	/* test tag name */

	if( toupper( (unsigned char) s[ 0 ] ) == 'S' &&
		toupper( (unsigned char) s[ 1 ] ) == 'C' &&
		toupper( (unsigned char) s[ 2 ] ) == 'R' &&
		toupper( (unsigned char) s[ 3 ] ) == 'I' &&
		toupper( (unsigned char) s[ 4 ] ) == 'P' &&
		toupper( (unsigned char) s[ 5 ] ) == 'T' &&
		(
			'\0' == s[ 6 ] ||
			'>'  == s[ 6 ] ||
			isspace( (unsigned char ) s[ 6 ] )
		) )
	{
		if( ending )
			in_script = 0;   /* it's </SCRIPT> */
		else
			in_script = 1;   /* it's <SCRIPT>  */
	}

	return;
}
