/*-------------------------------------------------
 Routine for scrubbing HTML tags.

 Copyright (C) 2004  Scott McKellar  mck9@swbell.net

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 02111-1307, USA.
 ------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "cfg.h"
#include "hscrub.h"

#define NAME_LEN_MAX CFG_TEXT_MAX

static char name_buff [ NAME_LEN_MAX + 1 ];
static char attr_name [ NAME_LEN_MAX + 1 ];
static Config config;
static unsigned dropped = 0;

static int edit_tag( const char * s, FILE * pOut, size_t line );
static const char * find_attribute( const char * s );

void install_config( Config cfg )
{
	config = cfg;
}

int scrub_tag( const char * s, FILE * pOut, size_t line )
{
	int rc = 0;
	const char * p = s;  /* for traversing string */
	int ending = 0;      /* boolean; set true for ending tag */
	size_t i;            /* index into name_buff */

	/* Null pointers mean to clean internal resources */
	/* and get out. */

	if( ! p || ! pOut )
	{
		dropped = 0;
		return 0;
	}

	/* skip leading white space */

	while( isspace( (unsigned char) *p ) )
		++p;

	if( '/' == *p )
	{
		/* virgule found; we have an end tag.  Skip white space. */

		ending = 1;
		do ++p; while( isspace( (unsigned char) *p ) );
	}

	/* copy up to the next white space character */

	for( i = 0;
		 i < NAME_LEN_MAX && *p && ! isspace( (unsigned char) *p );
		 ++i, ++p )
		 name_buff [ i ] = toupper( (unsigned char) *p );

	/* add terminal nul */

	name_buff [ i ] = '\0';

	/* branch on action type */

	switch( cfg_action( config, name_buff ) )
	{
		case CFGA_KEEP :
			if( 0 == dropped )
				fprintf( pOut, "<%s>", s );
			break;
		case CFGA_WARN :
		{
			const char * virgule = ending ? "/" : "";
			fprintf( stderr, "Warning: <%s%s> tag on line %lu\n",
				virgule, name_buff, (unsigned long) line );
			if( 0 == dropped )
				fprintf( pOut, "<%s>", s );
			break;
		}
		case CFGA_DROP :
			break;             /* do nothing */
		case CFGA_DROP_ALL :
			if( ending )
			{
				if( dropped > 0 )
					--dropped;
			}
			else
				++dropped;
			break;
		case CFGA_EDIT :
			if( 0 == dropped )
			{
				if( ending )
					fprintf( pOut, "<%s>", s );
				else
					rc = edit_tag( s, pOut, line );
			}
			break;
		default :
			fputs( "scrub_tag: invalid action\n", stderr );
			rc = 1;
			break;
	}

	return rc;
}

/***********************************************
 Write text without alteration -- unless we are
 in the middle of a DROP ALL section.
 **********************************************/
int scrub_text( const char * s, FILE * f, size_t line )
{
	if( ! s || ! f )
	{
		dropped = 0;
		return 0;
	}

	if( dropped > 0 )
		return 0;

	if( fputs( s, f ) == EOF )
	{
		perror( "Unable to write HTML" );
		return 1;
	}
	else
		return 0;
}

/**************************************************
 Apply the previously designated action to a comment
 --  unless we are in the middle of a DROP ALL section.
 *************************************************/
int scrub_comment( const char * s, FILE * f, size_t line )
{
	int rc = 0;

	if( ! s || ! f )
	{
		dropped = 0;
		return 0;
	}

	switch( cfg_comment_action( config ) )
	{
		case CFGA_KEEP :
			if( 0 == dropped )
				fprintf( f, "<%s>", s );
			break;
		case CFGA_WARN :
			fprintf( stderr, "Warning: comment on line %ld\n",
				(unsigned long) line );
			if( 0 == dropped )
				fprintf( f, "<%s>", s );
			break;
		case CFGA_DROP :
			break;
		default:
			fputs( "scrub_comment: invalid action\n", stderr );
			rc = 1;
			break;
	}

	return rc;
}

/********************************************
 Echo the tag, but look up each attribute in the configuration
 file and either echo it, suppress it, or issue a warning
 for it.
 *******************************************/
static int edit_tag( const char * s, FILE * pOut, size_t line )
{
	int rc = 0;
	const char * end_p;

	fputc( '<', pOut );

	/* echo any leading white space */

	while( isspace( (unsigned char) *s ) )
		fputc( *s++, pOut );

	/* echo the tag name */

	while( *s && ! isspace( (unsigned char) *s ) )
		fputc( *s++, pOut );

	while( *s )
	{
		end_p = find_attribute( s );
		if( attr_name[ 0 ] )
		{
			switch( cfg_attr_action( config, name_buff, attr_name ) )
			{
				case CFGA_KEEP :
					fwrite( s, 1, end_p -s, pOut );
					break;
				case CFGA_DROP :
					break;
				case CFGA_WARN :
					fprintf( stderr,
						"line %lu: %s tag with attribute %s\n",
						(unsigned long) line, name_buff, attr_name );
					fwrite( s, 1, end_p -s, pOut );
					break;
				default :
					fputs( "edit_tag: Internal error: "
						"invalid attribute action\n", stderr );
					fwrite( s, 1, end_p -s, pOut );
						break;
			}
		}
		else
			fputs( s, pOut );

		s = end_p;
	}

	fputc( '>', pOut );

	return rc;
}

/******************************************************
 Extract the next attribute name into attr_name[].
 Return a pointer to the character just past the end of
 the attribute.  In almost all cases that means one past
 the end of the attribute's value, but not all attributes
 have values.
 *****************************************************/
static const char * find_attribute( const char * s )
{
	/* skip leading white space */

	while( isspace( (unsigned char) *s ) )
		++s;

	if( ! *s )
	{
		/* Found nothing but white space.  Return */
		/* an empty attribute name and get out.   */

		attr_name[ 0 ] = '\0';
		return s;
	}

	/* copy attribute name into buffer, */
	/* while raising it to upper case   */

	{
		int i = 0;   /* subscript */

		for( i = 0; i < NAME_LEN_MAX; ++i )
		{
			if( '=' == *s
				|| '\0' == *s
				|| isspace( (unsigned char) *s ) )
			{
				attr_name[ i ] = '\0';
				break;
			}
			else
				attr_name[ i ] = toupper( (unsigned char ) *s++ );
		}
	}

	/* See if the attribute name is too long */

	if( *s && *s != '=' && ! isspace( (unsigned char) *s ) )
	{
		/* Skip over the rest of the attribute name. */
		/* Strictly speaking we may misinterpret the */
		/* truncated attribute name, but this        */
		/* possibility is so remote in practice that */
		/* we just won't worry about it. */

		do ++s;
		while( *s && *s != '=' && !isspace( (unsigned char) *s ) );
	}

	/* Find the end of the attribute, which is  */
	/* usually the end of the associated value. */
	/* But an attribute may not have a value.   */

	/* Skip white space */

	while( isspace( (unsigned char) *s ) )
		++s;

	if( '=' == *s )
	{
		/* Skip over equals sign & any additional white space */

		++s;
		while( isspace( (unsigned char) *s ) )
			++s;
	}
	else
	{
		/* No equal sign?  This attribute */
		/* has no value, so we're done.   */

		return s;
	}

	/* If we get this far, we have found an equals   */
	/* sign, so we can expect a value.  Find the end */
	/* of the value. */

	/* Skip white space */

	while( isspace( (unsigned char) *s ) )
		++s;

	if( ! *s )
	{
		/* Equals sign, but no value?  Hmmm... We're done. */

		return s;
	}
	else if( '\'' == *s || '\"' == *s )
	{
		/* Value is delimited by a double or single quote. */
		/* Look for the corresponding closing delimiter.   */

		int delim = *s;

		++s;
		while( *s != delim && *s )
			++s;

		/* Assuming we found the delimiter rather */
		/* than a terminal nul, skip over it.     */

		if( *s )
			++s;
	}
	else
	{
		/* Value is not quoted.  Look      */
		/* for white space as a delimiter. */

		while( *s && ! isspace( (unsigned char) *s ) )
			++s;
	}

	return s;
}
